--File name  : autoText.e
--Info       : Print text at graphic co-ordinates in the colour specified.
--           : The routine automatically detects the background colour
--           : under each pixel and so dose not disturb the background.
--           : 3D effect or easily achieved using print3Dxy().
--           : Text can be center aligned with printcxy and print3Dcxy().
--           : Shaded text can also be easily achieved.
--           : Text can be entered from the key board with getxy().
--           : Alterative fonts can be loaded in place of the default one. 
--Author     : Roy Shepherd 
--           : Version 2.00
--Creation   : 3 September 1998  
--Contact    : Bugs or anything else just E-mail me, sroy@globalnet.co.uk      

------------------------------------<fontloader>-----------------------------

--attempt to load default font file 'serif.fnt'. 
--if unsuccessful abort programme with a message 
--An alterative font can be loaded if desired
                  
--global function fontLoader(sequence fileName)
               
------------------------------<printxy>---------------------------------------
--displays text at graphic co-ordinates in colour chosen without disturbing
--any other graphics or background colours.
        
--global procedure printxy(atom x,      --x co-ordinate
--                         atom y,      --y co-ordinate   
--                         atom sx,     --scale x co-ordinate
--                         atom sy,     --scale y co-ordinate
--                         atom colour, --colour for text
--                         sequence text,  --text to be displayed
--                         integer shading)--if TRUE shading is on

------------------------------<print3Dxy>----------------------------------------
--calls printxy twice with the x & y co-ordinates slightly offset 
--the second time to give a 3D effect

--global procedure print3Dxy(atom x,         --x co-ordinates 
--                           atom y,         --y co-ordinates 
--                           atom sx,        --x scaling
--                           atom sy,        --y scaling
--                           atom colour1,   --bottom colour for 3D
--                           atom colour2,   --top colour for 3D
--                           sequence text,  --3D text to be displayed
--                           integer shading)-- if TRUE shading is on
                  
-----------------------------<printcxy>---------------------------------------
--works out where to start text for displaying in centre of screen
--then calls printxy to display it
               
--global procedure printcxy(atom y,          --y co-ordinate
--                          atom sx,         --x scaling
--                          atom sy,         --y scaling
--                          atom colour,     --text colour
--                          sequence text,   --text to be displayed
--                          integer shading) --if TRUE shading is on
               
               
-----------------------------<print3Dcxy>---------------------------------------
--works out where to display 3Dtext in centre of sceen
--then calls print3Dxy to display it
               
--global procedure print3Dcxy(atom y,       --y co-ordinates
--                            atom sx,      --x scale
--                            atom sy,      --y scale
--                            atom colour1, --bottom colour
--                            atom colour2, --top colour
--                            sequence text,--3D text to be displayed
--                            integer shading)--if TRUE shading is on

---------------------------------<getxy>-------------------------------------
--graphic mode input function

--global function getxy(
--              atom x,           --x co-ordinates
--              atom y,           --y co-ordinates      
--              atom sx,          --scale up x co-ordinates
--              atom sy,          --scale up y co-ordinates
--              atom min,         --minimum allowed chaacter
--              atom max,         --maximum allowed character
--              atom len,         --maximum allowed length of userString
--              atom forColour,   --foreground colour for text
--              sequence extras,     --extra chaacters not in mim to max
--              sequence startText,  --prompt text
--              sequence userString) --user string, can contine text at start


---------------------------------<scalyX>------------------------------------
--scale up the x co-ordinate to the required width
                                         
--function scaleX(sequence fontInfo,atom sx)
  
--------------------------<autoDetectBackground>------------------------------
--sets fontString's background colours to the same as those already 
--on the screen. Set the forground colour to the one chosen.
--Set shading up if chosen.
    
--function autoDetectBackground(
--                           sequence fontString,--font from the loaded file
--                           atom x,          -- x co-ordinate
--                           atom y,          -- y co-ordinate
--                           atom sy,         -- y scaling
--                           atom colour,     --forground colour
--                           integer shading) --if TRUE shading is on
  
-------------------------------------------------------------------------------  
--includes--
include graphics.e 
include image.e
include get.e

--local constants--          
constant FALSE     =  0
constant BACKSPACE =  8
constant RETURN    = 13
constant ESCAPE    = 27

--local variable--
object font     
                 
-------------------------------<fontloader>----------------------------------

--attempt to load default font file 'serif.fnt'. 
--if unsuccessful program will display a message and abort 
--An alterative font can be loaded if desired
                 
global procedure fontLoader(sequence fileName)
object fontLine
integer fn 
font={}
fn=open(fileName,"r")
if fn=-1 then
    if graphics_mode(-1) then end if
    puts(1,"AutoText Could,t open " & fileName)--display message
    abort(1)--failed to open a font file so abort with a message
end if                

-- if opened successfully read it in 
    while 1 do
        fontLine = gets(fn)
        if atom(fontLine) then
            exit
        end if  
        if fontLine[1]>'-' then--don't load comments
           font = append(font, fontLine[1..8])--remove carrage return
        end if
    end while        
    close(fn)--close file  
end procedure 
                                                     
---------------------------------<scalyX>------------------------------------
--scale up the x co-ordinate to the required width
                                         
function scaleX(sequence fontInfo,atom sx)
sequence scaleXup 
integer hight,inc
inc=1
hight=length(fontInfo)--get hight of fontInfo
scaleXup=repeat(0,hight)--set up veriable for scaling up          
for y=1 to hight do    
    scaleXup[y]=repeat(0,sx*8) 
    for x = 1 to 8 do                                                           
        for scale=1 to sx do                       
--fontInfo[y][x] go's in to scaleXup[y][x] scale number of times-increasing it's size
            scaleXup[y][inc+scale-1]=fontInfo[y][x]
        end for
        inc=inc+sx--increment inc scaleX times
    end for       
    inc=1--back to one ready for next row of font
end for      
return scaleXup--return scaled up fout to fontInfo
end function
            
--------------------------<autoDetectBackground>------------------------------
--sets fontString's background colours to the same as those already 
--on the screen. Set the forground colour to the one chosen.
--Set shading up if chosen.
    
function autoDetectBackground(sequence fontString,--font from the loaded file
                              atom x,          -- x co-ordinate
                              atom y,          -- y co-ordinate
                              atom sy,         -- y scaling
                              atom colour,     --forground colour
                              integer shading) --if TRUE shading is on
integer len,hight --length and hight of font   
sequence background      
len=length(fontString[1])--get length of font  
hight=length(fontString)--get hight of font
    background=save_image({x,y},{x+len,y+hight})--get image of background
    for yy=1 to hight do--loop through entire font changing forground
        for xx=1 to len do--and background pixels
            if fontString[yy][xx]='.' then        --marry foreground
               fontString[yy][xx]=background[yy][xx]--with background  
               else
                 fontString[yy][xx]=colour--set forground colour
             end if                                 
        end for      
       if shading=1 then colour=colour+1/sy end if--set shading if TRUE
    end for           
  return fontString--return ready to be displayed
end function     
       
------------------------------<printxy>---------------------------------------
--displays text at graphic co-ordinates in colour chosen without disturbing
--any other graphics or background colours.
        
global procedure printxy(atom x,      --x co-ordinate
                         atom y,      --y co-ordinate   
                         atom sx,     --scale x co-ordinate
                         atom sy,     --scale y co-ordinate
                         atom colour, --colour for text
                         sequence text,  --text to be displayed
                         integer shading)--if TRUE shading is on
                  

sequence fontInfo 
atom fontNum,len 
len=sx*8--calculate length font will be when scaled up
x=x-8*sx--adjust x co-ordinates
for textSt=1 to length(text) do
     fontInfo={}
     fontNum=text[textSt]*16+1        
     --read one character from text[] into fontInfo
     for read=fontNum to fontNum+15 do--read each line of font
         for scaleY = 1 to sy do--scale up y co-ordinates
             fontInfo = append(fontInfo,font[read])--read a line at a time 
        end for                                    --into fontInfo
     end for   
     if sx>1 then fontInfo=scaleX(fontInfo,sx) end if--scale up x co-ordinates
--marry foreground with background and set colour and shading if TRUE         
     fontInfo=autoDetectBackground(fontInfo,x+textSt*len,y,sy,colour,shading)
--display one character on the screen then loop back for next       
     display_image({x+textSt*len,y},fontInfo)
end for      
end procedure  

------------------------------<print3Dxy>----------------------------------------
--calls printxy twice with the x & y co-ordinates slightly offset 
--the second time to give a 3D effect

global procedure print3Dxy(atom x,         --x co-ordinates 
                           atom y,         --y co-ordinates 
                           atom sx,        --x scaling
                           atom sy,        --y scaling
                           atom colour1,   --bottom colour for 3D
                           atom colour2,   --top colour for 3D
                           sequence text,  --3D text to be displayed
                           integer shading)-- if TRUE shading is on
atom Dx,Dy
Dx=1 Dy=1
if sx > 1 then Dx=floor(sx/2) end if--calculate for 3D effect
if sy > 1 then Dy=floor(sy/2) end if                   
    printxy(x,y,sx,sy,colour1,text,shading)
    printxy(x+Dx,y+Dy,sx,sy,colour2,text,shading)--offset for 3D effect     
end procedure

-----------------------------<printcxy>---------------------------------------
--works out where to start text for displaying in centre of screen
--then calls printxy to display it
               
global procedure printcxy(atom y,          --y co-ordinate
                          atom sx,         --x scaling
                          atom sy,         --y scaling
                          atom colour,     --text colour
                          sequence text,   --text to be displayed
                          integer shading) --if TRUE shading is on
  sequence config    
  atom x,len
  len=sx*8--calculate what length of text will be when Xscaled
  config = video_config()
   x = config[5]--5 holds the width of the screen in pixels   
   x = x/2--take graphic mode used into account
   x = x - length(text)*(len)/2
   printxy(x,y,sx,sy,colour,text,shading)
end procedure
 
-----------------------------<print3Dcxy>---------------------------------------
--works out where to display 3Dtext in centre of sceen
--then calls print3Dxy to display it
               
global procedure print3Dcxy(atom y,       --y co-ordinates
                            atom sx,      --x scale
                            atom sy,      --y scale
                            atom colour1, --bottom colour
                            atom colour2, --top colour
                            sequence text,--3D text to be displayed
                            integer shading)--if TRUE shading is on
  sequence config    
  atom x,len
  len=sx*8--calculate what length of text will be when Xscaled 
  config = video_config()
   x = config[5]--5 holds the width of the screen in pixels   
   x = x/2--take graphic mode used into account
   x = x - length(text)*(len)/2
   print3Dxy(x,y,sx,sy,colour1,colour2,text,shading)
end procedure

---------------------------------<tickDelay>--------------------------------
--used for short delays when flasing cursor in getxy(). it's also used in beep()
--setting d to 1 will cause a delay of one tick of clock
procedure tickDelay(atom d)
    atom oldTime
    integer ticks
    oldTime=time()
    ticks=0
    while d>ticks do
          while oldTime=time() do end while 
        ticks=ticks+1
        oldTime=time()
    end while                            
end procedure

----------------------------------<beep>------------------------------------
--beep when user reaches maximum line length in getxy() routine

procedure beep() 
    sound(1000)
    tickDelay(3)--beep for three ticks of clock
    sound(0)
end procedure  

---------------------------------<getxy>-------------------------------------
--graphic mode input function that works with printxy

global function getxy(atom x,           --x co-ordinates
                      atom y,           --y co-ordinates      
                      atom sx,          --scale up x co-ordinates
                      atom sy,          --scale up y co-ordinates
                      atom min,         --minimum allowed chaacter
                      atom max,         --maximum allowed character
                      atom len,         --maximum allowed length of userString
                      atom forColour,   --foreground colour for text
                      sequence extras,     --extra chaacters not in mim to max
                      sequence startText,  --prompt text
                      sequence userString) --user string, can contine text at start

sequence ch,capture,cur
atom keyPress
atom count,xinc,yinc
--len = len + length(userString)
count=length(userString)*8*sx
len=len*8*sx--length in pixels of a character  
xinc=sx*8--calculate x increase           
yinc=sy*16-- calculate y increase         
printxy(x,y,sx,sy,forColour,startText,FALSE)--display prompt text if there is any
x=x+length(startText)*8*sx--advance x to end of prompt text  
capture=save_image({x,y},{x+len+sx*8,y+yinc})--save images of screen where
                                             -- text will be entered 
printxy(x,y,sx,sy,forColour,userString,FALSE)-- display any text already in
   while 1 do                                -- userString
      while 1 do
          keyPress=get_key()--look for key pressed 
          if keyPress !=-1 then exit end if--if key is pressed exit first loop 
          cur=get_pixel({x+count+sx*4,y+yinc-4,100})--no cursor in graphic mode
                                                    --so make one
          draw_line(forColour,{{x+count+sx*4,y+yinc-4},{x+count+xinc+sx*4,y+yinc-4}})
          tickDelay(2)--give cursor a blinking effect 
          pixel(cur,{x+count+sx*4,y+yinc-4})
          keyPress=get_key()--look for key pressed 
          if keyPress !=-1 then exit end if--if key pressed exit loop and process
          tickDelay(2)
      end while
      ch=sprintf("%s",keyPress)--put key pressed into ch
      --check that keypress is in range and userString < len  
      if keyPress >= min and keyPress <= max or match(ch,extras) then 
            if count < len then   
                 userString = userString & ch--put ch in userString
                 printxy(x,y,sx,sy,forColour,userString,FALSE)--display userString
                 count=count+xinc--increase count by one character
                 else beep()--userString at maximum so beep if any more character 
             end if         --are attempted to be entered
       end if
      if keyPress = BACKSPACE and count >0 then--backspace key as been pressedd
           count = count - xinc--decrease count one character
           userString=userString[1..length(userString)-1]--decrease userString by 1
           display_image({x,y},capture)--display the previously capture background
           printxy(x,y,sx,sy,forColour,userString,FALSE)--display userString            
                                                --without the last charactar 
     end if 
     --exit loop if either RETURN or ESCAPE is pressed
    if(keyPress=RETURN or keyPress=ESCAPE) then exit end if    

   end while                      
    if keyPress=ESCAPE then
       userString={}--if ESCAPE is pressed then return nothing 
    end if      
return userString--if RETURN is pressed return whatever is in userString    
end function

-------------------------------------------------------------------------------
 
fontLoader("serif.fnt")--default font 
 
     














